/* FlamingModule.cp by Abraham Flaxman (abie@kenlaw.com)
   Copyright (C) 1996

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */ 

// all code "borrowed" from BouncingBall by Hiroshi Lockheimer
// and Plasma by Christian Bauer

#include "FlamingModule.h"

typedef unsigned char UBYTE;
typedef unsigned long ULONG;

const int DISPLAY_X = 320;
const int DISPLAY_Y = 240;

BLocker		*gLock;
thread_id	gThreadID;

// Global variables
UBYTE *Buffer1, *Buffer2;
UBYTE *SrcBuffer, *DestBuffer;

UBYTE PalRed[256];
UBYTE PalGreen[256];
UBYTE PalBlue[256];

UBYTE ColorConv[256];

int xmod;

BBitmap *the_bitmap;
BView *the_view;
BPoint gSrcPoint;

void
module_initialize(
	void	*inSettings,
	long 	inSettingsSize)
{
#pragma unused (inSettings, inSettingsSize)
}

void
module_cleanup(
	void	**outSettings,
	long	*outSettingsSize)
{
	*outSettings = NULL;
	*outSettingsSize = 0;
}

void
module_start_saving(
	BView	*inView)
{
	srand((long)system_time());
	gLock = new BLocker();
	int i;

	the_view = inView;
	// Allocate bitmap and buffers
	the_bitmap = new BBitmap(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1), B_COLOR_8_BIT);
	Buffer1 = new UBYTE[DISPLAY_X * (DISPLAY_Y + 3)];
	Buffer2 = new UBYTE[DISPLAY_X * (DISPLAY_Y + 3)];
	memset(Buffer1, 0, DISPLAY_X * (DISPLAY_Y + 3));
	memset(Buffer2, 0, DISPLAY_X * (DISPLAY_Y + 3));

	xmod = the_bitmap->BytesPerRow();

	// Init buffer pointers
	SrcBuffer = Buffer1;
	DestBuffer = Buffer2;

	// Create color palette
	for (i=0; i<256; i++)
		PalRed[i] = PalGreen[i] = PalBlue[i] = 0;
	for (i=0; i<8; i++)		// 0..7: increase blue
		PalBlue[i] = i << 3;
	for (i=0; i<8; i++)		// 8..15: decrease blue
		PalBlue[i+8] = (7-i) << 3;
	for (i=8; i<32; i++)	// 8..31: increase red
		PalRed[i] = (i-8) * 255 / (32-8+1);
	for (i=32; i<56; i++) {	// 32..55: increase green
		PalRed[i] = 255;
		PalGreen[i] = (i-32) * 255 / (56-32+1);
	}
	for (i=56; i<80; i++) {	// 56..79: increase blue
		PalRed[i] = PalGreen[i] = 255;
		PalBlue[i] = (i-56) * 255 / (80-56+1);
	}
	for (i=80; i<256; i++)
		PalRed[i] = PalGreen[i] = PalBlue[i] = 255;

	// Create color conversion table
	for (i=0; i<256; i++)
		ColorConv[i] = index_for_color(PalRed[i], PalGreen[i], PalBlue[i]);
	
	if(the_view->Window()->Lock()) {
		gSrcPoint = the_view->Bounds().LeftTop()
						+the_view->Bounds().RightBottom()
						-BPoint(DISPLAY_X, DISPLAY_Y);
		gSrcPoint.x /= 2;
		gSrcPoint.y /= 2;
		
		the_view->Window()->Unlock();
	}

	gThreadID = spawn_thread(AnimMain, "Plasma",
							 B_DISPLAY_PRIORITY, NULL);
	resume_thread(gThreadID);
}

void
module_stop_saving()
{
	if (gLock->Lock()) {
		kill_thread(gThreadID);
				
		// Deallocate bitmap and buffers
		if (Buffer2)
			delete[] Buffer2;
		if (Buffer1)
			delete[] Buffer1;
		if (the_bitmap)
			delete the_bitmap;

		gLock->Unlock();
		delete (gLock);
	}
}

void
module_start_config(
	BView *inView)
{
	BRect bounds = inView->Bounds();
	float width = bounds.Width();
	float height = bounds.Height();
	
	BView *view = NULL;
	BStringView *str_view = NULL;

	view = new BView(
				BRect(10, 10, width-10, height-32),
					B_EMPTY_STRING,
					B_FOLLOW_ALL, 0);
	inView->AddChild(view);
	view->SetViewColor(inView->ViewColor());

	str_view = new BStringView(
				BRect(10, height-30, width-10, height-18 ),
				B_EMPTY_STRING,
				"FlamingModule, by Abraham Flaxman (abie@kenlaw.com)");
	str_view->SetAlignment(B_ALIGN_CENTER);
	inView->AddChild(str_view);
	str_view->SetViewColor(inView->ViewColor());

	str_view = new BStringView(
				BRect(10, height-16, width-10, height-4),
				B_EMPTY_STRING,
				"Based on Plasma by Christian Bauer");
	str_view->SetAlignment(B_ALIGN_CENTER);
	inView->AddChild(str_view);
	str_view->SetViewColor(inView->ViewColor());

	module_start_saving(view);
}

void
module_stop_config()
{
	module_stop_saving();
}

long
AnimMain(
	void	*data)
{
#pragma unused (data)

	while (TRUE) {
		UBYTE *p, *q, *r;
		int x, y;
		int decr;
	
		if(gLock->Lock()) {
			if(the_view->Window()->Lock()) {
				// Copy up by 1 pixel and interpolate
				p = SrcBuffer + 2 * DISPLAY_X; q = DestBuffer;
				decr = 1;
				for (y=0; y<DISPLAY_Y; y++) {
					for (x=0; x<DISPLAY_X; x++) {
						ULONG sum;
						sum = p[-DISPLAY_X-1];
						sum += p[-DISPLAY_X];
						sum += p[-DISPLAY_X+1];
						sum += p[-1];
						sum += p[1];
						sum += p[DISPLAY_X-1];
						sum += p[DISPLAY_X];
						sum += p[DISPLAY_X+1];
						sum >>= 3;
						if (sum)
							sum -= decr;	// Decrement every second row
						*q++ = sum;
						p++;
					}
					decr ^= 1;
				}
			
				// Fill lower 3 rows (invisible) with random pixels
				p = DestBuffer + DISPLAY_X * DISPLAY_Y + (DISPLAY_X-256) / 2;
				q = p + DISPLAY_X;
				r = q + DISPLAY_X;
				for (x=0; x<256; x++) {
					*p++ = (rand() & 0xf) + 0x40;
					*q++ = (rand() & 0xf) + 0x40;
					*r++ = (rand() & 0xf) + 0x40;
				}
			
				// Insert random bright spots
				q = DestBuffer + DISPLAY_X * (DISPLAY_Y+1) + (DISPLAY_X-256) / 2;
				int num_spots = rand() & 0xf;
				for (x=0; x<num_spots; x++) {
					p = q + (rand() & 0xff);
					p[-DISPLAY_X-1] = p[-DISPLAY_X] = p[-DISPLAY_X+1]
						= p[-1] = p[0] = p[1]
						= p[DISPLAY_X-1] = p[DISPLAY_X] = p[DISPLAY_X+1] = 255;
				}
			
				// Copy buffer to bitmap (and convert colors)
				p = DestBuffer;
				q = (UBYTE *)the_bitmap->Bits();
				for (y=0; y<DISPLAY_Y; y++) {
					for (x=0; x<DISPLAY_X; x++)
						q[x] = ColorConv[p[x]];
					p += DISPLAY_X;
					q += xmod;
				}
			
				// Swap buffer pointers
				UBYTE *tmp = SrcBuffer;
				SrcBuffer = DestBuffer;
				DestBuffer = tmp;
				
				the_view->DrawBitmap(the_bitmap, gSrcPoint);
				the_view->Window()->Unlock();
			}
		}
		gLock->Unlock();
		snooze(10000);		// 1/100th of a second	
	}
	
	return (0);
}
